#include "GLWindow.h"
#include <iostream>
#include "ngl/Camera.h"
#include "ngl/Light.h"
#include "ngl/Matrix.h"
#include "ngl/Transformation.h"
#include "ngl/TransformStack.h"
#include "ngl/Material.h"
#include "ngl/NGLInit.h"
#include "ngl/VBOPrimitives.h"
#include "ngl/ShaderManager.h"
#include "ngl/Random.h"
#include <boost/foreach.hpp>
#include "ngl/Util.h"

const static int s_extents=30;
//----------------------------------------------------------------------------------------------------------------------
GLWindow::GLWindow(
									 int _numSpheres,
                   QWidget *_parent
                  ) :
                    QGLWidget(_parent)
{
  // set this widget to have the initial keyboard focus
  setFocus();
  // re-size the widget to that of the parent (in this case the GLFrame passed in on construction)
  this->resize(_parent->size());
  // Now set the initial GLWindow attributes to default values
  // Roate is false
  m_rotate=false;
  // mouse rotation values set to 0
  m_spinXFace=0;
  m_spinYFace=0;
  m_origX=0;
  m_origY=0;
  m_sphereUpdateTimer=startTimer(40);
  m_animate=true;
	m_checkSphereSphere=true;
	// now create the actual spheres for our program
	m_bbox = new ngl::BBox(ngl::Vector(0,0,0),80,80,80);
	// create vectors for the position and direction
	m_numSpheres=_numSpheres;
  resetSpheres();

}


void GLWindow::resetSpheres()
{
	std::vector<Sphere>::iterator begin=m_sphereArray.begin();
	std::vector<Sphere>::iterator end=m_sphereArray.end();
  m_sphereArray.erase(begin,end);
	ngl::Vector pos(0,0,0);
  ngl::Vector dir;
	ngl::Random *rng=ngl::Random::instance();
	// loop and create the initial particle list
	for(int i=0; i<m_numSpheres; ++i)
	{
		dir=rng->getRandomVector();
		// add the spheres to the end of the particle list
		m_sphereArray.push_back(Sphere(rng->getRandomPoint(s_extents,s_extents,s_extents),dir,rng->randomPositiveNumber(2)+0.5));
	}

}


GLWindow::~GLWindow()
{
  ngl::NGLInit *Init = ngl::NGLInit::instance();
  std::cout<<"Shutting down NGL, removing VBO's and Shaders\n";
  Init->NGLQuit();
}

// This virtual function is called once before the first call to paintGL() or resizeGL(),
//and then once whenever the widget has been assigned a new QGLContext.
// This function should set up any required OpenGL context rendering flags, defining display lists, etc.

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::initializeGL()
{
  // create the lights

  glClearColor(0.4f, 0.4f, 0.4f, 1.0f);			   // Grey Background
  // enable depth testing for drawing
   glEnable(GL_DEPTH_TEST);
   // Now we will create a basic Camera from the graphics library
   // This is a static camera so it only needs to be set once
   // First create Values for the camera position
   ngl::Vector From(60,60,80);
   ngl::Vector To(0,0,0);
   ngl::Vector Up(0,1,0);
   ngl::NGLInit *Init = ngl::NGLInit::instance();
   #ifdef WIN32
   	glewInit();
   #endif
   Init->initGlew();
   m_cam= new ngl::Camera(From,To,Up,ngl::PERSPECTIVE);
   // set the shape using FOV 45 Aspect Ratio based on Width and Height
   // The final two are near and far clipping planes of 0.5 and 10
   m_cam->setShape(45,(float)720.0/576.0,0.5,150,ngl::PERSPECTIVE);
   // now to load the shader and set the values
   // grab an instance of shader manager
   ngl::ShaderManager *shader=ngl::ShaderManager::instance();
   // load a frag and vert shaders
   shader->loadShader("gl3xTest","shaders/Vertex.vs","shaders/Fragment.fs");
   // set this as the active shader
   shader->useShader("gl3xTest");
   // now pass the modelView and projection values to the shader
   shader->setShaderParamFromMatrix("gl3xTest","ViewMatrix",m_cam->getModelView());
   shader->setShaderParamFromMatrix("gl3xTest","projectionMatrix",m_cam->getProjection());
   shader->setShaderParam1i("gl3xTest","Normalize",true);
   shader->setShaderParam1i("gl3xTest","numLightsEnabled",1);

   glEnable(GL_DEPTH_TEST); // for removal of hidden surfaces

   ngl::VBOPrimitives *prim =  ngl::VBOPrimitives::instance();
   prim->createVBOSphere("sphere",1.0,40);
	 shader->loadShader("Colour","shaders/Colour.vs","shaders/Colour.fs");

		 // set this as the active shader
	   shader->useShader("Colour");
		 shader->setShaderParamFromMatrix("Colour","ViewMatrix",m_cam->getModelView());
	   shader->setShaderParamFromMatrix("Colour","projectionMatrix",m_cam->getProjection());

		 shader->setShaderParam4f("Colour","Colour",1,1,1,1);

}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget has been resized.
// The new size is passed in width and height.
void GLWindow::resizeGL(
                        int _w,
                        int _h
                       )
{
  glViewport(0,0,_w,_h);
  m_cam->setShape(45,(float)_w/_h,0.5,150,ngl::PERSPECTIVE);
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  (*shader)["gl3xTest"]->use();
	shader->setShaderParamFromMatrix("gl3xTest","projectionMatrix",m_cam->getProjection());
  (*shader)["Colour"]->use();
	shader->setShaderParamFromMatrix("Colour","projectionMatrix",m_cam->getProjection());

}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget needs to be painted.
// this is our main drawing routine
void GLWindow::paintGL()
{
  // clear the screen and depth buffer
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 // build our transform stack
  ngl::Transformation trans;
  // set the mouse rotation
  trans.setRotation(m_spinXFace,m_spinYFace,0);
  // set this in the TX stack
  m_transformStack.setGlobal(trans);

  // grab an instance of the primitives for drawing
  ngl::VBOPrimitives *prim=ngl::VBOPrimitives::instance();
  // Rotation based on the mouse position for our global
  // transform
  // grab an instance of the shader manager
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  (*shader)["Colour"]->use();
	m_transformStack.loadGlobalAndCurrentMatrixToShader("Colour","ModelMatrix");

	m_bbox->draw();

	shader->use("gl3xTest");
	ngl::Material m(ngl::GOLD);
	m.use();

	m_transformStack.loadGlobalAndCurrentMatrixToShader("gl3xTest","ModelMatrix");
  shader->setShaderParamFromMatrix("gl3xTest","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());
	BOOST_FOREACH(Sphere s, m_sphereArray)
	{
		s.draw("gl3xTest",m_transformStack);
	}

}




//----------------------------------------------------------------------------------------------------------------------
void GLWindow::update()
{
	BOOST_FOREACH(Sphere &s, m_sphereArray)
	{
		s.move();
	}
	checkCollisions();
}



//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseMoveEvent (
                               QMouseEvent * _event
                              )
{
  // note the method buttons() is the button state when event was called
  // this is different from button() which is used to check which button was
  // pressed when the mousePress/Release event is generated
  if(m_rotate && _event->buttons() == Qt::LeftButton)
  {
    m_spinYFace = ( m_spinYFace + (_event->x() - m_origX) ) % 360 ;
    m_spinXFace = ( m_spinXFace + (_event->y() - m_origY) ) % 360 ;
    m_origX = _event->x();
    m_origY = _event->y();
  }
  // re-draw GL
  updateGL();
}


//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mousePressEvent (
                                QMouseEvent * _event
                               )
{
  // this method is called when the mouse button is pressed in this case we
  // store the value where the maouse was clicked (x,y) and set the Rotate flag to true
  if(_event->button() == Qt::LeftButton)
  {
    m_origX = _event->x();
    m_origY = _event->y();
    m_rotate =true;
  }
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseReleaseEvent (
                                  QMouseEvent * _event
                                 )
{
  // this event is called when the mouse button is released
  // we then set Rotate to false
  if (_event->button() == Qt::LeftButton)
  {
    m_rotate=false;
  }
}

void GLWindow::timerEvent(
                          QTimerEvent *_event
                         )
{
	if(_event->timerId() == m_sphereUpdateTimer)
	{
		if (m_animate !=true)
		{
			return;
		}
		update();
		updateGL();
	}
}

void GLWindow::keyPress(QKeyEvent *_event)
{

  switch (_event->key())
  {
		case  Qt::Key_Space : m_animate^=true; break;
	  case Qt::Key_S : m_checkSphereSphere^=true; break;
		case Qt::Key_R : resetSpheres(); break;
	case Qt::Key_Minus : removeSphere(); break;
	case Qt::Key_Plus : addSphere(); break;
	}

}


bool GLWindow::sphereSphereCollision(
																		 ngl::Vector _pos1,
																		 GLfloat _radius1,
																		 ngl::Vector _pos2,
																		 GLfloat _radius2
																		)
{
  // the relative position of the spheres
  ngl::Vector relPos;
  //min an max distances of the spheres
  GLfloat dist;
  GLfloat minDist;
  GLfloat len;
  relPos =_pos1-_pos2;
  // and the distance
  len=relPos.length();
  dist=len*len;
  minDist =_radius1+_radius2;
  // if it is a hit
  if(dist <=(minDist * minDist))
  {
    return true;
  }
  else
  {
    return false;
  }
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::BBoxCollision()
{
  //create an array of the extents of the bounding box
  float ext[6];
  ext[0]=ext[1]=(m_bbox->height()/2.0f);
  ext[2]=ext[3]=(m_bbox->width()/2.0f);
  ext[4]=ext[5]=(m_bbox->depth()/2.0f);
  // Dot product needs a Vector so we convert The Point Temp into a Vector so we can
  // do a dot product on it
  ngl::Vector p;
  // D is the distance of the Agent from the Plane. If it is less than ext[i] then there is
  // no collision
  GLfloat D;
  // Loop for each sphere in the vector list
  BOOST_FOREACH(Sphere &s, m_sphereArray)
  {
    p=s.getPos();
    //Now we need to check the Sphere agains all 6 planes of the BBOx
    //If a collision is found we change the dir of the Sphere then Break
    for(int i=0; i<6; ++i)
    {
      //to calculate the distance we take the dotporduct of the Plane Normal
      //with the new point P
      D=m_bbox->getNormalArray()[i].dot(p);
      //Now Add the Radius of the sphere to the offsett
      D+=s.getRadius();
      // If this is greater or equal to the BBox extent /2 then there is a collision
      //So we calculate the Spheres new direction
      if(D >=ext[i])
      {
        //We use the same calculation as in raytracing to determine the
        // the new direction
        GLfloat x= 2*( s.getDirection().dot((m_bbox->getNormalArray()[i])));
        ngl::Vector d =m_bbox->getNormalArray()[i]*x;
        s.setDirection(s.getDirection()-d);
        s.setHit();
      }//end of hit test
     }//end of each face test
    }//end of for
}

void  GLWindow::checkSphereCollisions()
{
  bool collide;

  unsigned int size=m_sphereArray.size();

  for(unsigned int ToCheck=0; ToCheck<size; ++ToCheck)
	{
    for(unsigned int Current=0; Current<size; ++Current)
    {
      // don't check against self
      if(ToCheck == Current)  continue;

      else
      {
        //cout <<"doing check"<<endl;
        collide =sphereSphereCollision(m_sphereArray[Current].getPos(),m_sphereArray[Current].getRadius(),
                                       m_sphereArray[ToCheck].getPos(),m_sphereArray[ToCheck].getRadius()
                                      );
        if(collide== true)
        {
          m_sphereArray[Current].reverse();
          m_sphereArray[Current].setHit();
				}
      }
    }
  }
}


void  GLWindow::checkCollisions()
{

	if(m_checkSphereSphere == true)
	{
		checkSphereCollisions();
  }
	BBoxCollision();

}

void GLWindow::removeSphere()
{
	std::vector<Sphere>::iterator end=m_sphereArray.end();
  if(--m_numSpheres==0)
	{
		m_numSpheres=1;
	}
	else
	{
		m_sphereArray.erase(end-1,end);
	}
}

void GLWindow::addSphere()
{
	ngl::Random *rng=ngl::Random::instance();
	ngl::Vector dir;
	dir=rng->getRandomVector();
	// add the spheres to the end of the particle list
	m_sphereArray.push_back(Sphere(rng->getRandomPoint(s_extents,s_extents,s_extents),dir,rng->randomPositiveNumber(2)+0.5));
	++m_numSpheres;
}



